# -*- coding: UTF-8 -*-
import sys
import MySQLdb
import logging.handlers
import traceback
from datetime import datetime, date, timedelta

log = logging.getLogger('su.artix.db_trimmer')
handler = logging.handlers.RotatingFileHandler('log_db_trimmer.log', maxBytes=1000000, backupCount=10, delay=False)
handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)-8s %(message)s '))
log.addHandler(handler)
log.setLevel(logging.INFO)


specialPKs = {'client': 'clientitemid', 'couponitem': 'couponid', 'users': 'workshiftid', 'cardChanged': 'id', 'clientChanged': 'id', 'excisemarks': 'id'}
specialTimeFieldNames = {'cardChanged': 'date', 'clientChanged': 'date'}

# Таблицы, которые нельзя удалять напрямую по полю с датой и временем (У ЭТИХ ТАБЛИЦ НЕТ ИНДЕКСА ПО ПОЛЮ ДАТЫ И ВРЕМЕНИ)
delByIndexFieldTables = ['cardChanged', 'clientChanged']

class TableTrimmer(object):
    def __init__(self, cursor, tableName, pkName, timeFieldName):
        self.tableName = tableName
        self.pkName = pkName
        self.timeFieldName = timeFieldName
        self.minId = None
        self.minTime = None
        self.maxId = None
        self.maxTime = None
        self.cursor = cursor

    def setMinIdAndTime(self):
        query = 'SELECT %s, %s FROM %s WHERE %s = (SELECT MIN(%s) FROM %s)' % (self.pkName, self.timeFieldName, self.tableName, self.pkName, self.pkName, self.tableName)
        self.cursor.execute(query)
        if self.cursor.rowcount > 0:
            self.minId, self.minTime = self.cursor.fetchone()
        log.debug('%-15s: minId = %s, minTime = %s' % (self.tableName, self.minId, self.minTime))

    def setMaxIdAndTime(self):
        query = 'SELECT %s, %s FROM %s WHERE %s = (SELECT MAX(%s) FROM %s)' % (self.pkName, self.timeFieldName, self.tableName, self.pkName, self.pkName, self.tableName)
        self.cursor.execute(query)
        if self.cursor.rowcount > 0:
            self.maxId, self.maxTime = self.cursor.fetchone()
        log.debug('%-15s: maxId = %s, maxTime = %s' % (self.tableName, self.maxId, self.maxTime))

    def getNextExistsUpdateTime(self, idX):
        self.cursor.execute("SELECT %s FROM %s WHERE %s >= %s ORDER BY %s LIMIT 1" % (self.timeFieldName, self.tableName, self.pkName, idX, self.pkName))
        return self.cursor.fetchone()[0]


    def getId(self, minT, maxT, t, minId, maxId, k=0):
        log.debug('%-15s: minT = %s, maxT = %s, minId = %s, maxId = %s' % (self.tableName, minT, maxT, minId, maxId))
        idDeltaMain = maxId - minId
        timeDeltaMain = (maxT - minT).total_seconds()
        timeDeltaX = (maxT - t).total_seconds()
        idX = int(maxId - (idDeltaMain*timeDeltaX//timeDeltaMain))
        resultT = self.getNextExistsUpdateTime(idX)
        log.debug('%-15s: Дельта %s' % (self.tableName, (t-resultT).total_seconds()))
        if abs((t-resultT).total_seconds()) > 300 and k < 100:
            if resultT > t:
                return self.getId(minT, resultT, t, minId, idX, k + 1)
            else:
                return self.getId(resultT, maxT, t, idX, maxId, k + 1)

        if resultT > t:
            log.info('%-15s: Удалим записи с %s < %s' % (self.tableName, self.timeFieldName, resultT))
            return idX-1
        log.info('%-15s: Удалим записи с %s <= %s' % (self.tableName, self.timeFieldName, resultT))
        return idX

    # удалить записи раньше указанной даты
    def trim(self, actualDate):
        # если можно выполнить запрос сразу указав до какого времени записи можно удалить, то так и сделаем
        if not self.tableName in delByIndexFieldTables:
            query = 'DELETE FROM %s WHERE %s <= "%s"' % (self.tableName, self.timeFieldName, actualDate)
            log.info('%-15s: %s' % (self.tableName, query))
            self.cursor.execute(query)
            log.info('%-15s: Удалено %s записей' % (self.tableName, self.cursor.rowcount))
            return

        self.setMinIdAndTime()
        self.setMaxIdAndTime()
        searchTime = datetime.strptime(actualDate, '%Y-%m-%d %H:%M:%S')
        if not self.minId:
            log.info('%-15s: В таблице нет ни одной записи.' % self.tableName)
        elif not self.minTime or not self.maxTime or self.minTime > self.maxTime:
            log.warning('%-15s: В таблице некорректные данные. Выполнять удаление не будем.' % self.tableName)
        elif searchTime > self.maxTime:
            log.info('%-15s: Все данные устарели. Выполним очистку таблицы.' % self.tableName)
            query = 'TRUNCATE TABLE %s' % self.tableName
            log.info('%-15s: %s' % (self.tableName, query))
            self.cursor.execute(query)
        elif searchTime < self.minTime:
            log.info('%-15s: Все данные актуальны.' % self.tableName)
        elif searchTime > self.minTime and searchTime < self.maxTime:
            deleteFromId = self.getId(self.minTime, self.maxTime, searchTime, self.minId, self.maxId)
            query = 'DELETE FROM %s WHERE %s <= %s' % (self.tableName, self.pkName, deleteFromId)
            log.info('%-15s: %s' % (self.tableName, query))
            self.cursor.execute(query)
            log.info('%-15s: Удалено %s записей' % (self.tableName, self.cursor.rowcount))


def trim(host, port, user, passwd, db, tables, actual_days):
    try:
        conn = MySQLdb.connect(host=host, port=port, user=user, passwd=passwd, db=db)
        conn.set_character_set('utf8')
        conn.autocommit(True)
    except MySQLdb.Error as e:
        log.error("Error %d: %s" % (e.args[0], e.args[1]))
        sys.exit (1)
    log.info('host=%s, port=%s, db=%s, actual_days=%s' % (host, port, db, actual_days))
    cursor = conn.cursor()
    start = datetime.now()
    for t in tables:
        if t:
            pk = specialPKs.get(t, '%sid' % t)
            timeFieldName = specialTimeFieldNames.get(t, 'update_time')
            trimmer = TableTrimmer(cursor, t, pk, timeFieldName)
            try:
                trimmer.trim((datetime.now() - timedelta(days=actual_days)).strftime('%Y-%m-%d 00:00:00'))
            except Exception:
                log.error(traceback.format_exc())
    log.info('Выполнено за %s' % (datetime.now() - start))

    conn.close()
